--- ========================================================= ---
---  Setup Revision, Version, Date and add to AceLibrary
--- ========================================================= ---
local VERSION = tonumber(("$Revision: 71480 $"):match("%d+"))

local Sphere = Sphere
Sphere_Buttons = Sphere:NewModule("Buttons", "AceEvent-2.0", "AceHook-2.1")
local self = Sphere_Buttons
if Sphere.revision < VERSION then
	Sphere.version = "r" .. VERSION
	Sphere.revision = VERSION
	Sphere.date = ("$Date: 2008-04-25 12:54:12 -0400 (Fri, 25 Apr 2008) $"):match("%d%d%d%d%-%d%d%-%d%d")
end
Sphere.desc = "Adds Buttons to the Sphere"

--- ========================================================= ---
---  Setup Locals
--- ========================================================= ---
G = AceLibrary('Gratuity-2.0')

local L = Sphere:L("Sphere_Buttons")

local data = {
	activeButtons = {},
	unactiveButtons = {},
	maxSpellID = nil,
	cursorType = nil,
	cursorInfo1 = nil,
	cursorInfo2 = nil,
	cursorHasItem = false,
}

local menuTypeValidate = {
	autohide = L["Auto Hide"],
	always = L["Always"],
	toggle = L["Toggle"],
}								

local menuTypeValidateDesc = {}

local menuButtonValidate = {
	left = L["Left Mouse Button"],
	right = L["Right Mouse Button"],
	middle = L["Middle Mouse Button"],
}								

local menuButtonValidateDesc = {}

local buttonTypeValidate = {
	none = L["Normal"],
	stance = L["Stance/Stealth"],
	harmful = L["Harmful/Helpful"],
	modifier = L["Modifier"],
}

local buttonTypeValidateDesc = {
	none = L["Selecting this will have your Button and buttons associated with it to respond normally.  Changing the Type will cause the Button and all buttons associated with it to be deleted."],
	stance = L["Selecting this will have your Button and buttons associated with it change depending on your stance/stealth.  Changing the Type will cause the Button and all buttons associated with it to be deleted."],
	harmful = L["Selecting this will have your Button and buttons associated with it change depending on if your target is friendly or hostile.  Changing the Type will cause the Button and all buttons associated with it to be deleted."],
	modifier = L["Selecting this will have your Button and buttons associated with it change depending on a modifier (shift, alt, ctrl).  Changing the Type will cause the Button and all buttons associated with it to be deleted."],
}

local buttonUnitValidate = {
	player = L["Player"],
	target = L["Target"],
	targettarget = L["Target of Target"],
	focus = L["Focus"],
	pet = L["Pet"],
	party1 = L["Party Member1"],
	party2 = L["Party Member2"],
	party3 = L["Party Member3"],
	party4 = L["Party Member4"],
	none = L["Normal"],
}

local buttonUnitValidateDesc = {}

local buttonShownValidate = {
	always = L["Always"],
	never = L["Never"],
	combat = L["Combat"],
	nocombat = L["Not in Combat"],
	unit = L["Unit"],
}

local buttonShownValidateDesc = {}

--- ========================================================= ---
---  On Intitalize / Setup Database
--- ========================================================= ---
function Sphere_Buttons:OnInitialize()
	self.db = Sphere:AcquireDBNamespace("Buttons")
	Sphere:RegisterDefaults("Buttons", "char", {
		settings = {
			scale = 0.85,
			offSet = 60,
			menuType = "always",
			menuButton = {
				left = false,
				middle = false,
				right = true,
			},
		},
		colors = {
			oor = {0.75, 0.10, 0.10, 1},
			oom = {0.10, 0.25, 1, 1},
			unusable = {0.50, 0.50, 0.50, 1},
			itemequippedborder = {0.10, 1, 0.10, 1},
			buttonhighlightborder = {0, 0.25, 1, 1},
		},
		
		SphereBar1 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar2 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar3 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar4 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar5 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar6 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar7 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar8 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar9 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar10 = {
			menu = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
	})
	
	data.maxSpellID = self:FindMaxID()
	Sphere_Buttons:CreateFrames()
end

--- ========================================================= ---
---  Setup Options
--- ========================================================= ---
Sphere_Buttons:RegisterSphereOption({
	name = L["Buttons"],
	desc = L["Change the Buttons options."],
	type = 'group',
	args = {
		appearance = {
			name = L["Appearance"],
			desc = L["Change how the Buttons Appearance looks in general."],
			type = 'group',
			args = {
				colors = {
					name = L["Colors"],
					desc = L["Change the Colors that the Buttons use."],
					type = 'group',
					pass = true,
					get = function(value)
						return unpack(self.db.char.colors[value])
					end,
					set = "SetColor",
					args = {
						oor = {
							name = L["Out of Range"],
							desc = L["Color for when Out of Range."],
							type = 'color',
							hasAlpha = false,
						},
						oom = {
							name = L["Out of Mana"],
							desc = L["Color for when Out of Mana."],
							type = 'color',
							hasAlpha = false,
						},
						unusable = {
							name = L["Unusable"],
							desc = L["Color for when Unusable."],
							type = 'color',
							hasAlpha = false,
						},
						itemequippedborder = {
							name = L["Item Equipped Border"],
							desc = L["Color for your Item Equipped Border."],
							type = 'color',
							hasAlpha = false,
						},
						buttonhighlightborder = {
							name = L["Button Highlight Border"],
							desc = L["Color for you mouseover a Button."],
							type = 'color',
							hasAlpha = false,
						},
					},
				},
				size = {
					name = L["Size"],
					desc = L["Set how large the Buttons are."],
					type = 'range',
					min = 0.50,
					max = 2,
					step = 0.01,
					bigStep = 0.05,
					get = function()
						return self.db.char.settings.scale
					end,
					set = "SetScale"
				},
				offSet = {
					name = L["offSet"],
					desc = L["Set how much of an offSet the Buttons are."],
					type = 'range',
					min = 50,
					max = 150,
					step = 1,
					bigStep = 5,
					get = function()
						return self.db.char.settings.offSet
					end,
					set = "SetOffSet"
				},
			},
		},
		settings = {
			name = L["Settings"],
			desc = L["Change the Button Settings in general."],
			type = 'group',
			args = {
				type = {
					name = L["Menu Type"],
					desc = L["Change how the Menu operates."],
					type = 'text',
					validate = menuTypeValidate,
					validateDesc = menuTypeValidateDesc,
					get = function()
						return self.db.char.settings.menuType
					end,
					set = "SetMenuType",
				},
				button = {
					name = L["Menu Button"],
					desc = L["Change how the Menu operates."],
					type = 'text',
					multiToggle = true,
					validate = menuButtonValidate,
					validateDesc = menuButtonValidateDesc,
					get = function(key)
						return self.db.char.settings.menuButton[key]
					end,
					set = function(key, value) 
						self.db.char.settings.menuButton[key] = value;
						self:SetMenuButton();
					end,
				},
			},
		},			
		button1 = {
			name = L["Button 01"],
			desc = L["Change Button 1 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar1.menu
					end,
					set = function() 
						self.db.char.SphereBar1.menu = not(self.db.char.SphereBar1.menu);
						self:SetButtonToMenu("1")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar1.type
					end,
					set = function(name) 
						self:SetButtonType(name, "1")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar1.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "1")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar1.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "1")
					end,
				},
			},
		},
		button2 = {
			name = L["Button 02"],
			desc = L["Change Button 2 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar2.menu
					end,
					set = function() 
						self.db.char.SphereBar2.menu = not(self.db.char.SphereBar2.menu);
						self:SetButtonToMenu("2")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar2.type
					end,
					set = function(name) 
						self:SetButtonType(name, "2")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar2.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "2")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar2.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "2")
					end,
				},
			},
		},
		button3 = {
			name = L["Button 03"],
			desc = L["Change Button 3 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar3.menu
					end,
					set = function() 
						self.db.char.SphereBar3.menu = not(self.db.char.SphereBar3.menu);
						self:SetButtonToMenu("3")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar3.type
					end,
					set = function(name) 
						self:SetButtonType(name, "3")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar3.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "3")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar3.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "3")
					end,
				},
			},
		},
		button4 = {
			name = L["Button 04"],
			desc = L["Change Button 4 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar4.menu
					end,
					set = function() 
						self.db.char.SphereBar4.menu = not(self.db.char.SphereBar4.menu);
						self:SetButtonToMenu("4")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar4.type
					end,
					set = function(name) 
						self:SetButtonType(name, "4")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar4.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "4")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar4.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "4")
					end,
				},
			},
		},
		button5 = {
			name = L["Button 05"],
			desc = L["Change Button 5 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar5.menu
					end,
					set = function() 
						self.db.char.SphereBar5.menu = not(self.db.char.SphereBar5.menu);
						self:SetButtonToMenu("5")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar5.type
					end,
					set = function(name) 
						self:SetButtonType(name, "5")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar5.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "5")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar5.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "5")
					end,
				},
			},
		},
		button6 = {
			name = L["Button 06"],
			desc = L["Change Button 6 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar6.menu
					end,
					set = function() 
						self.db.char.SphereBar6.menu = not(self.db.char.SphereBar6.menu);
						self:SetButtonToMenu("6")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar6.type
					end,
					set = function(name) 
						self:SetButtonType(name, "6")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar6.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "6")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar6.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "6")
					end,
				},
			},
		},
		button7 = {
			name = L["Button 07"],
			desc = L["Change Button 7 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar7.menu
					end,
					set = function() 
						self.db.char.SphereBar7.menu = not(self.db.char.SphereBar7.menu);
						self:SetButtonToMenu("7")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar7.type
					end,
					set = function(name) 
						self:SetButtonType(name, "7")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar7.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "7")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar7.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "7")
					end,
				},
			},
		},
		button8 = {
			name = L["Button 08"],
			desc = L["Change Button 8 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar8.menu
					end,
					set = function() 
						self.db.char.SphereBar8.menu = not(self.db.char.SphereBar8.menu);
						self:SetButtonToMenu("8")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar8.type
					end,
					set = function(name) 
						self:SetButtonType(name, "8")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar8.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "8")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar8.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "8")
					end,
				},
			},
		},
		button9 = {
			name = L["Button 09"],
			desc = L["Change Button 9 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar9.menu
					end,
					set = function() 
						self.db.char.SphereBar9.menu = not(self.db.char.SphereBar9.menu);
						self:SetButtonToMenu("9")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar9.type
					end,
					set = function(name) 
						self:SetButtonType(name, "9")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar9.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "9")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar9.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "9")
					end,
				},
			},
		},
		button10 = {
			name = L["Button 10"],
			desc = L["Change Button 10 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar10.menu
					end,
					set = function() 
						self.db.char.SphereBar10.menu = not(self.db.char.SphereBar10.menu);
						self:SetButtonToMenu("10")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar10.type
					end,
					set = function(name) 
						self:SetButtonType(name, "10")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar10.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "10")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar10.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "10")
					end,
				},
			},
		},

	}

})

--- ========================================================= ---
---  On Enable
--- ========================================================= ---
function Sphere_Buttons:OnEnable()
	data.maxSpellID = self:FindMaxID()
	
	self:ScheduleRepeatingEvent(self.UpdateButton, TOOLTIP_UPDATE_TIME, self)
	self:ScheduleRepeatingEvent(self.FlashButton, ATTACK_BUTTON_FLASH_TIME, self)
	
	self:ScheduleEvent("Sphere_Buttons_PostEnable", self.OnPostEnable, 1, self)
end

--- ========================================================= ---
---  On Disable
--- ========================================================= ---
function Sphere_Buttons:OnDisable()
	self:UnregisterAllEvents();
end

--- ========================================================= ---
---  On Post Enable
--- ========================================================= ---
function Sphere_Buttons:OnPostEnable()
	Sphere_Buttons:ProcessActiveButtons()
	Sphere_Buttons:ProcessUnactiveButtons()
end

--- ========================================================= ---
--- Find Max Spell ID
--- ========================================================= ---
function Sphere_Buttons:FindMaxID()
	local id = 10
	while GetSpellTexture(id, BOOKTYPE_SPELL) do
		id = id + 1
	end
	return id
end

--- ========================================================= ---
--- Spell Name TO Spell ID
--- ========================================================= ---
function Sphere_Buttons:GetSpellInfo(spellName)
	for i = data.maxSpellID, 1, -1 do
		if (GetSpellName(i, BOOKTYPE_SPELL)) == spellName then
			return i
		end
	end
end

--- ========================================================= ---
--- Find Item Location
--- ========================================================= ---
function Sphere_Buttons:FindItemLocation(itemName)
	local b, s, i = nil;
	local iBag, iSlot = nil;
	local iInv = nil;
	local itemLoc, itemA, itemB = nil;
	
	for b=0,4 do
		for s=1,GetContainerNumSlots(b) do
			if GetContainerItemLink(b,s) then
				if string.find(GetContainerItemLink(b,s),itemName) then
					iBag = b
					iSlot = s
				end
			end
		end
	end
	
	if ((iBag == nil) and (iSlot == nil)) then
		for i=0,23 do
			if GetInventoryItemLink("player",i) then
				if string.find(GetInventoryItemLink("player",i),itemName) then
					iInv = i
				end
			end
		end
	end
	
	if ((iBag) and (iSlot)) then
		itemLoc = "Container"
		itemA = iBag
		itemB = iSlot
	elseif (iInv) then
		itemLoc = "Inventory"
		itemA = iInv
	end

	return itemLoc, itemA, itemB
end

--- ========================================================= ---
---  Set Color
--- ========================================================= ---
function Sphere_Buttons:SetColor(given_kind, r, g, b, a)
	local kind = given_kind
	local colors = self.db.char.colors[kind]
	
	if r then
		colors[1] = r
		colors[2] = g
		colors[3] = b
		colors[4] = a
	else
		r, g, b, a = unpack(colors)
	end
	
	if kind == "buttonhighlightborder" then
		for i=1, 10 do
			highlight = getglobal("SphereButton"..i):GetHighlightTexture();
			local r, g, b, a = unpack(self.db.char.colors.buttonhighlightborder)
			highlight:SetVertexColor(r, g, b, a)
		end
	end
	
end

--- ========================================================= ---
---  Set Scale
--- ========================================================= ---
function Sphere_Buttons:SetScale(value)
	if value then
		self.db.char.settings.scale = value
	else
		value = self.db.char.settings.scale
	end
	
	for i=1, 10 do
		getglobal("SphereBar"..i):SetScale(value)
	end
end

--- ========================================================= ---
---  Set offSet
--- ========================================================= ---
function Sphere_Buttons:SetOffSet(value)
	if value then
		self.db.char.settings.offSet = value
	else
		value = self.db.char.settings.offSet
	end
	
	for i=1, 10 do
		getglobal("SphereBar"..i):ClearAllPoints()
		local radians = -math.rad(((i-1) * 36) - 90);
		local xLoc = value * math.cos(radians)
		local yLoc = value * math.sin(radians)
		getglobal("SphereBar"..i):SetPoint("CENTER", xLoc, yLoc);
	end
end

--- ========================================================= ---
---  Set Menu Type
--- ========================================================= ---
function Sphere_Buttons:SetMenuType(value)
	if value then
		self.db.char.settings.menuType = value
	else
		value = self.db.char.settings.menuType
	end
	for i = 1, 10 do 
		local hdr = getglobal("SphereBar"..i.."Header")
		if (value == "autohide") then
			hdr:SetAttribute('delaytimemap-anchor', '3');
			hdr:SetAttribute('state', 0);
		elseif (value ==  "toggle") then
			hdr:SetAttribute('delaytimemap-anchor', "");
			hdr:SetAttribute('state', 0);
		elseif (value ==  "always") then
			hdr:SetAttribute('delaytimemap-anchor', "");
			if (self.db.char["SphereBar"..i].menu) then
				hdr:SetAttribute('state', 1);
			else
				hdr:SetAttribute('state', 0);
			end
		end
	end
end

--- ========================================================= ---
---  Set Menu Button
--- ========================================================= ---
function Sphere_Buttons:SetMenuButton()
	for i = 1, 10 do 
		local btn = getglobal("SphereBar"..i.."Button1")
		
		btn:SetAttribute('onmousedownbutton1', "");
		btn:SetAttribute('onmouseupbutton1', "");
		btn:SetAttribute('onmousedownbutton2', "");
		btn:SetAttribute('onmouseupbutton2', "");
		btn:SetAttribute('onmousedownbutton3', "");
		btn:SetAttribute('onmouseupbutton3', "");
		
		if btn:GetAttribute("*type1") == "" then
			btn:SetAttribute("*type1", nil)
		end
		if btn:GetAttribute("*type2") == "" then
			btn:SetAttribute("*type2", nil)
		end
		if btn:GetAttribute("*type3") == "" then
			btn:SetAttribute("*type3", nil)
		end

		if (self.db.char["SphereBar"..i].menu) then
			if (self.db.char.settings.menuButton.left) then
				btn:SetAttribute('onmousedownbutton1', 'buttondown');
				btn:SetAttribute('onmouseupbutton1', 'buttonup');
				
				btn:SetAttribute("*type1", ATTRIBUTE_NOOP)
			end
			if (self.db.char.settings.menuButton.right) then
				btn:SetAttribute('onmousedownbutton2', 'buttondown');
				btn:SetAttribute('onmouseupbutton2', 'buttonup');
				
				btn:SetAttribute("*type2", ATTRIBUTE_NOOP)
			end
			if (self.db.char.settings.menuButton.middle) then
				btn:SetAttribute('onmousedownbutton3', 'buttondown');
				btn:SetAttribute('onmouseupbutton3', 'buttonup');
				
				btn:SetAttribute("*type3", ATTRIBUTE_NOOP)
			end
		end
	end
end

--- ========================================================= ---
---  Set Button To Menu
--- ========================================================= ---
function Sphere_Buttons:SetButtonToMenu(id)
	if id then
		Sphere_Buttons:SetIcon(getglobal("SphereBar"..id.."Button1"))
	end
end

--- ========================================================= ---
---  Get Class Stance Set
--- ========================================================= ---
function Sphere_Buttons:GetClassStanceSet(englishClass)
	local statemaps
	if englishClass == "ROGUE" then
		statemaps = {
			["stealth:0"] = "s0",
			["stealth:1"] = "s1",
		}
	elseif englishClass == "WARRIOR" then
		statemaps = {
			["stance:1"] = "s1",
			["stance:2"] = "s2",
			["stance:3"] = "s3",
		}
	elseif englishClass == "PRIEST" then
		statemaps = {
			["stance:0"] = "s0",
			["stance:1"] = "s1",
		}
	elseif englishClass == "DRUID" then
		statemaps = {
			["stance:0"] = "s0",
			["stance:1"] = "s1",
			["stance:2"] = "s2",
			["stance:3"] = "s3",
			["stance:3, stealth"] = "st3",
			["stance:4"] = "s4",
			["stance:5"] = "s5",
			["stance:6"] = "s6",
		}
	end
	
	return statemaps
end

--- ========================================================= ---
---  Set Button Type
--- ========================================================= ---
function Sphere_Buttons:SetButtonType(value, id)
	if value then
		self.db.char["SphereBar"..id].type = value
	elseif id then
		value = self.db.char["SphereBar"..id].type
	end
	
	Sphere_Buttons:DeleteBar(id)

	if value == "harmful" then
		self.db.char["SphereBar"..id].useunit = 1
		getglobal("SphereBar"..id):SetAttribute("useunit", 1)
	else
		self.db.char["SphereBar"..id].useunit = nil
		getglobal("SphereBar"..id):SetAttribute("useunit", nil)
	end
	
	if value == "modifier" then
		self.db.char["SphereBar"..id].statemaps = {
			["modifier:shift"] = "s",
			["nomodifier"] = "n",
			["modifier:ctrl"] = "c",
			["modifier:alt"] = "a",
		}
		Sphere_Buttons:UpdateBar(getglobal("SphereBar"..id))
		if Sphere:HasModule('SubButtons') then
			Sphere_Buttons:UpdateBar(getglobal("SphereSubBar"..id))
		end
	elseif value == "stance" then
		local playerClass, englishClass = UnitClass("player");
		local statemaps = Sphere_Buttons:GetClassStanceSet(englishClass)
		self.db.char["SphereBar"..id].statemaps = statemaps
		Sphere_Buttons:UpdateBar(getglobal("SphereBar"..id))
		if Sphere:HasModule('SubButtons') then
			Sphere_Buttons:UpdateBar(getglobal("SphereSubBar"..id))
		end
	else
		self.db.char["SphereBar"..id].statemaps = {}
		Sphere_Buttons:UpdateBar(getglobal("SphereBar"..id))
		if Sphere:HasModule('SubButtons') then
			Sphere_Buttons:UpdateBar(getglobal("SphereSubBar"..id))
		end
	end
end

--- ========================================================= ---
---  Set Button Unit
--- ========================================================= ---
function Sphere_Buttons:SetButtonUnit(value, id)
	if value then
		self.db.char["SphereBar"..id].unit = value
	elseif id then
		value = self.db.char["SphereBar"..id].unit
	end
	
	Sphere_Buttons:UpdateBar(getglobal("SphereBar"..id))
	if Sphere:HasModule('SubButtons') then
		Sphere_Buttons:UpdateBar(getglobal("SphereSubBar"..id))
	end
end

--- ========================================================= ---
---  Set Button Shown
--- ========================================================= ---
function Sphere_Buttons:SetButtonShown(value, id)
	if value then
		self.db.char["SphereBar"..id].shown = value
	elseif id then
		value = self.db.char["SphereBar"..id].shown
	end
	
	Sphere_Buttons:UpdateBar(getglobal("SphereBar"..id))
	if Sphere:HasModule('SubButtons') then
		Sphere_Buttons:UpdateBar(getglobal("SphereSubBar"..id))
	end
end

--- ========================================================= ---
---  Process Active Buttons
--- ========================================================= ---
function Sphere_Buttons:ProcessActiveButtons()
	for i=1, #data.activeButtons do
		Sphere_Buttons:SetIcon(data.activeButtons[i])
	end
end

--- ========================================================= ---
---  Process Unactive Buttons
--- ========================================================= ---
function Sphere_Buttons:ProcessUnactiveButtons()
	for i=1, #data.unactiveButtons do
		Sphere_Buttons:SetIcon(data.unactiveButtons[i])
	end
end

--- ========================================================= ---
---  Create Frames
--- ========================================================= ---
function Sphere_Buttons:CreateFrames()
	for i=1, 10 do
		local bar = Sphere_Buttons:CreateBar(i, "SphereBar"..i, "SphereMainFrame", "SphereBar"..i, "Buttons")
		local btn = Sphere_Buttons:CreateButton(1, "SphereBar"..i, "SphereMainFrame", "SphereBar"..i, "SphereBar"..i, "Buttons")
		local hdr = Sphere_Buttons:CreateHeader(i, btn)
		
		bar = Sphere_Buttons:UpdateBar(bar)
		
		btn:SetAttribute('childraise1', true);
        btn:SetAttribute('childstate1', '$input');
		btn:SetAttribute('anchorchild-buttondown', hdr);
		btn:SetAttribute('childstate-buttondown', '^mousedown');
		btn:SetAttribute('childreparent-buttondown', true);
		btn:SetAttribute('anchorchild-buttonup', hdr);
		btn:SetAttribute('childstate-buttonup', 'mouseup');
		btn:SetAttribute('childverify-buttonup', true);
		btn:SetAttribute('onmouseupbutton1', '');
		
		bar:SetAttribute("addchild", btn);
	end
	Sphere_Buttons:SetMenuButton()
	Sphere_Buttons:SetMenuType()
end

--- ========================================================= ---
---  Create Header
--- ========================================================= ---
function Sphere_Buttons:CreateHeader(id, btn)
	local hdr = CreateFrame("Frame", "SphereBar"..id.."Header", btn, "SecureStateHeaderTemplate, SecureStateDriverTemplate")
	hdr:SetAttribute("checkselfcast", true)
	hdr:SetAttribute("useparent-unit", true)
	hdr:SetAttribute("useparent-statebutton", true)
	hdr:SetAllPoints(btn);
		
	hdr:SetAttribute('statemap-anchor-mousedown', '0-1');
	hdr:SetAttribute('statemap-anchor-mouseup', '1:');
	hdr:SetAttribute('delaystatemap-anchor', '0');
	hdr:SetAttribute('delaytimemap-anchor', "");
	hdr:SetAttribute('delayhovermap-anchor', 'true');
	
	return hdr
end

--- ========================================================= ---
---  Create Bar
--- ========================================================= ---
function Sphere_Buttons:CreateBar(id, name, anchor, valname, valdb)
	local offSet = self.db.char.settings.offSet
	local radians = -math.rad(((id-1) * 36) - 90);
	local xLoc = offSet * math.cos(radians)
	local yLoc = offSet * math.sin(radians)

	local bar = CreateFrame("Frame", name, getglobal(anchor), "SecureStateHeaderTemplate, SecureStateDriverTemplate")
	
	bar:SetID(id);
	bar:SetPoint("CENTER", xLoc, yLoc);
	bar:SetScale(self.db.char.settings.scale)
	bar.valname = valname
	bar.valdb = valdb

	return bar
end

--- ========================================================= ---
--- Create Button
--- ========================================================= ---
function Sphere_Buttons:CreateButton(id, name, anchor, baranchor, valname, valdb)
	local btn = CreateFrame("CheckButton", name.."Button"..id, getglobal(anchor), "SecureActionButtonTemplate, SecureAnchorUpDownTemplate, ActionButtonTemplate")
	btn = Sphere_Buttons:CreateButtonEvents(btn)
	
	btn:SetID(id)
	btn:SetAllPoints(baranchor);
	btn.valname = valname
	btn.valdb = valdb

	Sphere.db.namespaces[valdb].char[name].buttons[id] = Sphere.db.namespaces[valdb].char[name].buttons[id] or {}
	for k, v in pairs(Sphere.db.namespaces[valdb].char[name].buttons[id]) do
		btn:SetAttribute(k, v)
	end
	
	table.insert(data.activeButtons, btn)
	
	return btn
end

--- ========================================================= ---
--- Update Bar
--- ========================================================= ---
function Sphere_Buttons:UpdateBar(bar)
	local state
	local statelist, driver = {}, {}, {}

	local vals = Sphere.db.namespaces[bar.valdb].char[bar.valname]
	if vals.unit == "none" then
		bar:SetAttribute("unit", "")
	else
		bar:SetAttribute("unit", vals.unit)
	end
	bar:SetAttribute("shown", vals.shown)
	bar:SetAttribute("useunit", vals.useunit)
	UnregisterUnitWatch(bar)
	if vals.shown == "unit" then
		RegisterUnitWatch(bar)
	end
	if not vals.shown or vals.shown == "always" or (vals.shown == "nocombat" and not InCombatLockdown()) then
		bar:Show()
	else
		bar:Hide()
	end

	if vals.statemaps then
		for k, v in pairs(vals.statemaps) do
			if string.find(k, "%-") then
				k = string.gsub(k, "%-", ":")
			end

			local add = true
			local list = v..":"..v
			table.insert(driver, "["..k.."] "..v)

			for _, l in ipairs(statelist) do
				if l == list then
					add = nil
					break
				end
			end

			if add then
				table.insert(statelist, v..":"..v)
			end
		end
	end

	driver = table.concat(driver, ";")
	state = SecureCmdOptionParse(driver)
	statelist = table.concat(statelist, ";")

	if #driver == 0 then
		state = "0"
	end

	RegisterStateDriver(bar, "state", driver)
	bar:SetAttribute("statemap-state", "$input")
	bar:SetAttribute("statebutton", statelist)
	bar:SetAttribute("state", state)
	bar:SetWidth(35)
	bar:SetHeight(35)
	
	Sphere_Buttons:ProcessActiveButtons()
	Sphere_Buttons:ProcessUnactiveButtons()
	
	return bar
end

--- ========================================================= ---
---  Delete Bar
--- ========================================================= ---
function Sphere_Buttons:DeleteBar(id)
	local bar = getglobal("SphereBar"..id)
	local btn = getglobal("SphereBar"..id.."Button1")

	local attribs = Sphere.db.namespaces[btn.valdb].char[btn.valname].buttons[btn:GetID()]
	for k in pairs(attribs) do
		btn:SetAttribute(k, nil)
	end
	
	if Sphere:HasModule('SubButtons') then
		for i=1, 10 do
			local btn = getglobal("SphereSubBar"..id.."Button"..i)
			local attribs = Sphere.db.namespaces[btn.valdb].char[btn.valname].buttons[btn:GetID()]
			for k in pairs(attribs) do
				btn:SetAttribute(k, nil)
			end
		end
	end
end

--- ========================================================= ---
--- Convert Button
--- ========================================================= ---
function Sphere_Buttons:ConvertButton(btn, value)
	local unit = SecureButton_GetAttribute(btn, "unit")
	if unit and unit ~= "none" and UnitExists(unit) then
		if UnitCanAttack("player", unit) then
			value = SecureButton_GetModifiedAttribute(btn, "harmbutton", value) or value
		elseif UnitCanAssist("player", unit) then
			value = SecureButton_GetModifiedAttribute(btn, "helpbutton", value) or value
		end
	end
	return value
end

--- ========================================================= ---
--- Flash Button
--- ========================================================= ---
function Sphere_Buttons:FlashButton()
	for index=1, #data.activeButtons do
		local btn = data.activeButtons[index]
		local flash = getglobal(btn:GetName().."Flash")
		if btn.flash and btn.attack then
			if flash:IsVisible() then
				flash:Hide()
			else
				flash:Show()
			end
		else
			flash:Hide()
		end
	end
end

--- ========================================================= ---
--- Set Icon
--- ========================================================= ---
function Sphere_Buttons:SetIcon(btn)
	local value = btn:GetAttribute("state-parent")
	if value then
		value = Sphere_Buttons:ConvertButton(btn, value)
		local texture, spell
		local new = SecureButton_GetModifiedAttribute(btn, "type", value)
		btn.type = new
		btn.attack = nil
		btn.macro = nil
		if new then
			spell = SecureButton_GetModifiedAttribute(btn, new, value)
			if spell then
				if new == "spell" then
					texture = GetSpellTexture(spell)
					if IsAttackSpell(spell) or IsAutoRepeatSpell(spell) then
						btn.attack = true
					end
				elseif new == "item" then
					local itemid = SecureButton_GetModifiedAttribute(btn, "itemid", value)
					if itemid then
						texture = select(10, GetItemInfo(itemid))
					end
				elseif new == "macro" then
					btn.macro = spell
					spell, texture = GetMacroInfo(spell)
				end
			end
			btn.action = spell
		end
		
		local icon = getglobal(btn:GetName().."Icon")
		if icon:GetTexture() ~= texture then
			icon:SetTexture(texture)
			if texture then
				btn:Show()
				UIFrameFadeIn(icon, 0.50)
			end
			Sphere_Buttons:UpdateItem(btn)
			Sphere_Buttons:UpdateCooldown(btn)
		end
		
		if not texture and not btn.grid then
			for k, v in ipairs(data.activeButtons) do
				if v == btn then
					table.remove(data.activeButtons, k)
					table.insert(data.unactiveButtons, btn)
					break
				end
			end
		end
		
		if btn.valdb == "Buttons" then
			if Sphere.db.namespaces[btn.valdb].char[btn.valname].menu then
				btn:Show()
			else
				if not texture then
					btn:Hide()
				else
					btn:Show()
				end
			end
		else
			if not texture then
				btn:Hide()
			else
				btn:Show()
			end
		end

	end
end

--- ========================================================= ---
--- Update Button
--- ========================================================= ---
function Sphere_Buttons:UpdateButton()
	for index=1, #data.activeButtons do
		local btn = data.activeButtons[index]
		local r, g, b, a = 1, 1, 1, 1
		local unit = SecureButton_GetUnit(btn)
		if btn.type == "spell" then
			local usable, nomana = IsUsableSpell(btn.action) -- need check here
			if nomana then
				r, g, b, a = unpack(self.db.char.colors.oom)
			elseif not usable then
				r, g, b, a = unpack(self.db.char.colors.unusable)
			elseif SpellHasRange(btn.action) and IsSpellInRange(btn.action, unit) == 0 then
				r, g, b, a = unpack(self.db.char.colors.oor)
			end
		elseif btn.type == "item" then
			if not IsUsableItem(btn.action) then
				r, g, b, a = unpack(self.db.char.colors.unusable)
			elseif ItemHasRange(btn.action) and IsItemInRange(btn.action, unit) == 0 then
				r, g, b, a = unpack(self.db.char.colors.oor)
			end
		elseif btn.type == "macro" and GetMacroBody(btn.action) then
-- 			local action = GetActionFromMacroText(GetMacroBody(btn.action))
-- 			if action then
-- 				if GetItemInfo(action) then
-- 					if not IsUsableItem(action) then
-- 						r, g, b, a = unpack(self.db.char.colors.unusable)
-- 					elseif ItemHasRange(action) and IsItemInRange(action, unit) == 0 then
-- 						r, g, b, a = unpack(self.db.char.colors.oor)
-- 					end
-- 				else
-- 					local usable, nomana = IsUsableSpell(action)
-- 					if nomana then
-- 						r, g, b, a = unpack(self.db.char.colors.oom)
-- 					elseif not usable then
-- 						r, g, b, a = unpack(self.db.char.colors.unusable)
-- 					elseif SpellHasRange(action) and IsSpellInRange(action, unit) == 0 then
-- 						r, g, b, a = unpack(self.db.char.colors.oor)
-- 					end
-- 				end
-- 			end
			Sphere_Buttons:SetIcon(btn)
		end
		getglobal(btn:GetName().."Icon"):SetVertexColor(r, g, b, a)
	end
end

--- ========================================================= ---
--- Update Item
--- ========================================================= ---
function Sphere_Buttons:UpdateItem(btn)
	local count = getglobal(btn:GetName().."Count")
	local border = getglobal(btn:GetName().."Border")
	local num = nil
	if btn.action then
		if btn.type == "spell" then
			local spellname = string.gsub(btn.action, "(%w+[%s()%w]*)%(%)", "%1")
			local spellid = Sphere_Buttons:GetSpellInfo(spellname)
			G:SetSpell(spellid, "spell")
			local reg
			local reagent = G:Find(SPELL_REAGENTS, 2, 99, false, true, false) and true
			if reagent then
				for i = 1, G:NumLines() do
					if (G:Find(SPELL_REAGENTS, i, i, false, true, false) and true) then
						reg = G:GetLine(i)
						reg= string.gsub(reg, SPELL_REAGENTS, "")
						if string.find(reg, "^|c") then
							_, _, reg = string.find(reg, "|c%x+(.*)|r")
						end
						reg = string.gsub(reg, "%(%)", "")
					end
				end
			end
			if reg then
				num = GetItemCount(reg)
			end
		elseif btn.type == "item" then
			num = GetItemCount(btn.action)
		end
	end
	if num and num > 1 then
		count:SetText(num)
	else
		count:SetText(nil)
	end
end

--- ========================================================= ---
--- Update Cooldown
--- ========================================================= ---
function Sphere_Buttons:UpdateCooldown(btn)
	local start, duration, enable = 0, 0, 0
	local cooldown = getglobal(btn:GetName().."Cooldown")
	if btn.action then
		if btn.type == "spell" then
			start, duration, enable = GetSpellCooldown(btn.action)
		elseif btn.type == "item" then
			start, duration, enable = GetItemCooldown(btn.action)
		elseif btn.type == "macro" then
-- 			local action = GetActionFromMacroText(GetMacroBody(btn.action))
-- 			if action then
-- 				if GetItemInfo(action) then
-- 					start, duration, enable = GetItemCooldown(action)
-- 				else
-- 					start, duration, enable = GetSpellCooldown(action)
-- 				end
-- 			end
		end
	end
	if start and duration then
		CooldownFrame_SetTimer(cooldown, start, duration, enable)
	end
end

--- ========================================================= ---
--- Create Button Events
--- ========================================================= ---
function Sphere_Buttons:CreateButtonEvents(btn)
	btn:RegisterForClicks("AnyUp")
	btn:RegisterForDrag("LeftButton", "MiddleButton", "RightButton")
	btn:SetAttribute("checkselfcast", true)
	btn:SetAttribute("useparent-unit", true)
	btn:SetAttribute("useparent-statebutton", true)

	btn:SetScript("OnEnter", function(self)
			Sphere_Buttons:ButtonEnter(self) 
		end)
		
	btn:SetScript("OnLeave", function(self) 
			Sphere_Buttons:ButtonLeave(self) 
		end)
	
	btn:SetScript("OnDragStart", function(self)
			if not InCombatLockdown() then
			
				local clickType = 1;
				if (arg1 == "RightButton") then
					clickType = 2;
				end
				if (arg1 == "MiddleButton") then
					clickType = 3;
				end
				Sphere_Buttons:ButtonDragStart(self, clickType)
			end
		end)
	
	btn:SetScript("OnReceiveDrag", function() 
		end)
	
	btn:SetScript("PreClick", function(self)
			if GetCursorInfo() and not InCombatLockdown() then
				self:SetScript("OnClick", nil)
				
				local clickType = 1;
				if (arg1 == "RightButton") then
					clickType = 2;
				end
				if (arg1 == "MiddleButton") then
					clickType = 3;
				end
				
				local type, info1, info2
				
				if GetCursorInfo() then
					type, info1, info2 = GetCursorInfo()
				end
				Sphere_Buttons:ButtonReceiveDrag(self, clickType, type, info1, info2)
			end
		end)
		
	btn:SetScript("PostClick", function(self)
			self:SetChecked(nil)
			if not self:GetScript("OnClick") then
				self:GetParent():SetAttribute("addchild", self)
			end
		end)
		
	btn:SetScript("OnAttributeChanged", function(self, name, value)
			Sphere_Buttons:SaveAttrib(self, name, value)
		end)
	
	btn:SetScript("OnEvent", Sphere_Buttons.OnEvent);
	btn:RegisterEvent("PLAYER_ENTERING_WORLD");
	
	local cooldown = getglobal(btn:GetName().."Cooldown")
	cooldown:SetFrameLevel(cooldown:GetParent():GetFrameLevel() + 3)
	cooldown.text = cooldown:CreateFontString(nil, "OVERLAY")
	cooldown.text:SetPoint("CENTER")
	cooldown.text:SetFont("Fonts\\FRIZQT__.TTF", 22, "OUTLINE")
	
	local flash = getglobal(btn:GetName().."Flash")
	flash:ClearAllPoints()
	flash:SetAllPoints(cooldown)
	btn.flashtime = 0

	return btn
end

--- ========================================================= ---
--- Button Enter
--- ========================================================= ---
function Sphere_Buttons:ButtonEnter(this)
	if GetCursorInfo() then
		local type, info1, info2 = GetCursorInfo()
		data.cursorHasItem = true
		data.cursorType =  type
		data.cursorInfo1 =  info1
		data.cursorInfo2 =  info2
	else
		data.cursorType =  nil
		data.cursorInfo1 =  nil
		data.cursorInfo2 =  nil
	end
	
	if Sphere:HasModule('Tooltip') then
		if Sphere:IsModuleActive('Tooltip') then
	 		Sphere_Tooltip:SetTooltip(this)
		end
	end
end

--- ========================================================= ---
--- Button Leave
--- ========================================================= ---
function Sphere_Buttons:ButtonLeave(this)
	if not InCombatLockdown() and not Sphere.db.char.settings.locked then
		if (not(GetCursorInfo()) and (data.cursorHasItem)) then
			Sphere_Buttons:ButtonReceiveDrag(this, 2, data.cursorType, data.cursorInfo1, data.cursorInfo2)
			data.cursorHasItem = false
		end
	end

	GameTooltip:Hide() 
end

--- ========================================================= ---
--- Button Drag Start
--- ========================================================= ---
function Sphere_Buttons:ButtonDragStart(this, clickType)
	if not Sphere.db.char.settings.locked then
		local type, action = nil
		local bartype = Sphere.db.namespaces[this:GetParent().valdb].char[this:GetParent().valname].type
		local vals = Sphere.db.namespaces[this.valdb].char[this.valname].buttons
		
		if bartype == "stance" then
		
		elseif bartype == "modifiers" then
		
		elseif bartype == "harmful" then

		else
			if vals[this:GetID()]["*type"..clickType] then
				type = vals[this:GetID()]["*type"..clickType]
				vals[this:GetID()]["*type"..clickType] = nil
				if clickType == 1 then
					vals[this:GetID()]["*type*"] = nil
					this:SetAttribute("*type*", nil)
				end
			end
			
			if type then
				action = vals[this:GetID()]["*"..type..""..clickType]
				vals[this:GetID()]["*"..type..""..clickType] = nil
				if clickType == 1 then
					vals[this:GetID()]["*"..type.."*"] = nil
					this:SetAttribute("*"..action.."*", nil)
				end
			end
			
			if type == "spell" then
				PickupSpell(action)
			elseif type == "item" then
				local itemLoc, itemA, itemB = Sphere_Buttons:FindItemLocation(action)
				if itemLoc == "Container" then
					PickupContainerItem(itemA, itemB)
				elseif itemLoc == "Inventory" then
					PickupInventoryItem(itemA)
				end
			elseif type == "macro" then
				PickupMacro(action)
			end
		end
	end
end

--- ========================================================= ---
--- Button Drag Stop
--- ========================================================= ---
function Sphere_Buttons:ButtonReceiveDrag(this, clickType, type, info1, info2)
	if not Sphere.db.char.settings.locked then
		local state = this:GetAttribute("state-parent")
		if state == "0" then
			state = nil
		end
		
		if state then
			state = "-"..state
		else
			state = "*"
		end 
		
		local unit = SecureButton_GetModifiedUnit(this, (state or "LeftButton"))
		local useunit = this:GetParent():GetAttribute("useunit")
		if useunit and unit and unit ~= "none" and UnitExists(unit) then
			if UnitCanAttack("player", unit) then
				this:SetAttribute("*harmbutton"..state, "nuke"..state)
				state = "-nuke"..state
			elseif UnitCanAssist("player", unit) then
				this:SetAttribute("*helpbutton"..state, "heal"..state)
				state = "-heal"..state
			end
		end
	
		if clickType == 1 then
			this:SetAttribute("*type*", type)
			if type == "spell" then
				local spell, rank = GetSpellName(info1, info2)
				spell = spell.."()"
				this:SetAttribute("*spell"..state, spell)
				this:SetAttribute("*item"..state, nil)
				this:SetAttribute("*itemid"..state, nil)
				this:SetAttribute("*macro"..state, nil)
			elseif type == "item" then
				local item = GetItemInfo(info1)
				this:SetAttribute("*spell"..state, nil)
				this:SetAttribute("*item"..state, item)
				this:SetAttribute("*itemid"..state, info1)
				this:SetAttribute("*macro"..state, nil)
			elseif type == "macro" then
				this:SetAttribute("*spell"..state, nil)
				this:SetAttribute("*item"..state, nil)
				this:SetAttribute("*itemid"..state, nil)
				this:SetAttribute("*macro"..state, info1)
			end
		end

		local value = this:GetAttribute("state-parent")
		if value == "0" then
			value = nil
		end
		
		if value then
			value = "-"..value
		else
			value = clickType
		end
		
		local unit = SecureButton_GetModifiedUnit(this, (state or "LeftButton"))
		local useunit = this:GetParent():GetAttribute("useunit")
		if useunit and unit and unit ~= "none" and UnitExists(unit) then
			if UnitCanAttack("player", unit) then
				this:SetAttribute("*harmbutton"..value, "nuke"..value)
				value = "-nuke"..value
			elseif UnitCanAssist("player", unit) then
				this:SetAttribute("*helpbutton"..value, "heal"..value)
				value = "-heal"..value
			end
		end

		this:SetAttribute("*type"..value, type)
		if type == "spell" then
			local spell, rank = GetSpellName(info1, info2)
			spell = spell.."()"
			this:SetAttribute("*spell"..value, spell)
			this:SetAttribute("*item"..value, nil)
			this:SetAttribute("*itemid"..value, nil)
			this:SetAttribute("*macro"..value, nil)
		elseif type == "item" then
			local item = GetItemInfo(info1)
			this:SetAttribute("*spell"..value, nil)
			this:SetAttribute("*item"..value, item)
			this:SetAttribute("*itemid"..value, info1)
			this:SetAttribute("*macro"..value, nil)
		elseif type == "macro" then
			this:SetAttribute("*spell"..value, nil)
			this:SetAttribute("*item"..value, nil)
			this:SetAttribute("*itemid"..value, nil)
			this:SetAttribute("*macro"..value, info1)
		end

		for k, v in ipairs(data.unactiveButtons) do
			if v == this then
				table.remove(data.unactiveButtons, k)
				break
			end
		end
		table.insert(data.activeButtons, this)
		Sphere_Buttons:SetMenuButton()
		ClearCursor()
		
		data.cursorHasItem = false
		data.cursoractionType =  nil
		data.cursorInfo1 =  nil
		data.cursorInfo2 =  nil
	end
end

--- ========================================================= ---
--- Save Attributes
--- ========================================================= ---
function Sphere_Buttons:SaveAttrib(self, name, value)
	if string.find(name, "^%*") or string.find(name, "^shift") or string.find(name, "^ctrl") or string.find(name, "^alt") then
		Sphere.db.namespaces[self.valdb].char[self.valname].buttons[self:GetID()][name] = value
		name = "state-parent"
	end

	if name == "state-parent" then
		Sphere_Buttons:SetIcon(self)
	end
end

--- ========================================================= ---
--- Button OnEvent
--- ========================================================= ---
function Sphere_Buttons:OnEvent()
	if (event == "PLAYER_ENTERING_WORLD") then
		Sphere_Buttons:RegisterButtonEvents(this);
		return;
	end
	
	if (event == "PLAYER_REGEN_ENABLED") then
		local show = this:GetParent():GetAttribute("shown")
		if show == "nocombat" then
			this:GetParent():Show()
		elseif show == "combat" then
			this:GetParent():Hide()
		end
	elseif (event == "PLAYER_REGEN_DISABLED") then
		local show = this:GetParent():GetAttribute("shown")
		if show == "nocombat" then
			this:GetParent():Hide()
		elseif show == "combat" then
			this:GetParent():Show()
		end
	elseif (event == "PLAYER_TARGET_CHANGED") then
		if SecureButton_GetUnit(this) == "target" then
			Sphere_Buttons:SetIcon(this)
		end
	elseif (event == "PLAYER_FOCUS_CHANGED") then
		if SecureButton_GetUnit(this) == "focus" then
			Sphere_Buttons:SetIcon(this)
		end
	elseif (event == "MODIFIER_STATE_CHANGED") then
		Sphere_Buttons:SetIcon(this)
	elseif (event == "ACTIONBAR_UPDATE_USABLE") then
		Sphere_Buttons:UpdateCooldown(this)
	elseif (event == "UPDATE_INVENTORY_ALERTS") then
		Sphere_Buttons:UpdateCooldown(this)
	elseif (event == "ACTIONBAR_UPDATE_COOLDOWN") then
		Sphere_Buttons:UpdateCooldown(this)
	elseif (event == "UPDATE_MACROS") then
		Sphere_Buttons:SetIcon(this)
	elseif (event == "BAG_UPDATE") then
		Sphere_Buttons:UpdateItem(this)
	elseif (event == "UNIT_SPELLCAST_SUCCEEDED") then
		if this.type == "macro" then
			this.macro = true
		end
	elseif (event == "ACTIONBAR_SHOWGRID") then
		if not Sphere.db.char.settings.locked then
			this.grid = true
			if this.valdb ~= "Buttons" and Sphere.db.namespaces[this:GetParent().valdb].char[this:GetParent().valname].menu then
				local hdr = getglobal(this:GetParent():GetParent():GetName())
				if not hdr.oldstate then
					local oldstate = hdr:GetAttribute('state')
					hdr.oldstate = oldstate
					hdr:SetAttribute('state', 1);
				end
				this:Show()
			elseif this.valdb == "Buttons" then
				this:Show()
			else
				this:Hide()
			end
		end
	elseif (event == "ACTIONBAR_HIDEGRID") then
		if not Sphere.db.char.settings.locked then
			this.grid = nil
			if this.valdb ~= "Buttons" and Sphere.db.namespaces[this:GetParent().valdb].char[this:GetParent().valname].menu then
				local hdr = getglobal(this:GetParent():GetParent():GetName())
				if hdr.oldstate then
					hdr:SetAttribute('state', hdr.oldstate);
					hdr.oldstate = nil
				end
			end	
			if not getglobal(this:GetName().."Icon"):GetTexture() then
				if this.valdb == "Buttons" and Sphere.db.namespaces[this.valdb].char[this.valname].menu then
					this:Show()
				else
					this:Hide()
				end
			else
				this:Show()
			end
		end
	elseif (event == "PLAYER_ENTER_COMBAT") then
		this.flash = true
	elseif (event == "START_AUTOREPEAT_SPELL") then
		this.flash = true
	elseif (event == "PLAYER_LEAVE_COMBAT") then
		this.flash = nil
	elseif (event == "STOP_AUTOREPEAT_SPELL") then
		this.flash = nil
	elseif (event == "UNIT_FACTION") then
		if SecureButton_GetUnit(this) == arg1 then
			Sphere_Buttons:SetIcon(this)
		end
	end
end

--- ========================================================= ---
--- Register Button Events
--- ========================================================= ---
function Sphere_Buttons:RegisterButtonEvents(this)
	this:RegisterEvent("PLAYER_REGEN_ENABLED");
	this:RegisterEvent("PLAYER_REGEN_DISABLED");
	this:RegisterEvent("PLAYER_TARGET_CHANGED");
	this:RegisterEvent("PLAYER_FOCUS_CHANGED");
	this:RegisterEvent("MODIFIER_STATE_CHANGED");
	this:RegisterEvent("ACTIONBAR_UPDATE_USABLE");
	this:RegisterEvent("UPDATE_INVENTORY_ALERTS");
	this:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN");
	this:RegisterEvent("UPDATE_MACROS");
	this:RegisterEvent("BAG_UPDATE");
	this:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
	this:RegisterEvent("ACTIONBAR_SHOWGRID");
	this:RegisterEvent("ACTIONBAR_HIDEGRID");
	this:RegisterEvent("PLAYER_ENTER_COMBAT");
	this:RegisterEvent("START_AUTOREPEAT_SPELL");
	this:RegisterEvent("PLAYER_LEAVE_COMBAT");
	this:RegisterEvent("STOP_AUTOREPEAT_SPELL");
	this:RegisterEvent("UNIT_FACTION");
end