--- ========================================================= ---
---  Setup Revision, Version, Date and add to AceLibrary
--- ========================================================= ---
local VERSION = tonumber(("$Revision: 44649 $"):match("%d+"))

local Sphere = Sphere
Sphere_InnerGauge = Sphere:NewModule("InnerGauge", "AceEvent-2.0")
local self = Sphere_InnerGauge
if Sphere.revision < VERSION then
	Sphere.version = "r" .. VERSION
	Sphere.revision = VERSION
	Sphere.date = ("$Date: 2007-07-23 15:18:51 -0400 (Mon, 23 Jul 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end
Sphere.desc = "Adds a Inner Gauge to the Sphere"


--- ========================================================= ---
---  Setup Locals
--- ========================================================= ---
local DogTag = AceLibrary("DogTag-1.0")

local L = Sphere:L("Sphere_InnerGauge")

local start = nil

local eventType = {
	health = "[PercentHP]",
	power = "[PercentMP]",
	threat = "[PercentThreat]",
	experience = "[PercentXP]",
	reputation = "[PercentRep]",
	none = "",
	custom = "",
}

--- ========================================================= ---
---  On Intitalize / Setup Database
--- ========================================================= ---
function Sphere_InnerGauge:OnInitialize()
	self.db = Sphere:AcquireDBNamespace("InnerGauge")
	Sphere:RegisterDefaults("InnerGauge", "char", {
		gaugeUnit = "player",
		gaugeType = "power",
		gaugeCustom = "[PercentMP]",
		gaugeMarks = "four",
		gaugeMarksDarken = true,
		animate = true,
		useCustom = false,
	})

	self:CreateFrames()
	self:SetGaugeMarks()
	self:SetGaugeMarksDarken()
	
	if not Sphere:IsModuleActive('InnerGauge') then
		Sphere_InnerGauge:OnDisable()
	end
end

--- ========================================================= ---
---  Setup Validates for Options
--- ========================================================= ---
Sphere_InnerGauge_eventTypeValidate = {
	health = L["Health"],
	power = L["Power"],
	threat = L["Threat"],
	experience = L["Experience"],
	reputation = L["Reputation"],
	custom = L["Custom"],
	none = L["None"],
}

local eventTypeValidateDesc = {}

local eventUnitValidate = {
	player = L["Player"],
	target = L["Target"],
	focus = L["Focus"],
	pet = L["Pet"],
}

local eventUnitValidateDesc = {}

local gaugeMarksValidate = {
	none = L["None"],
	four = L["Four"],
	five = L["Five"],
	ten = L["Ten"],
	twenty = L["Twenty"],
}

local gaugeMarksValidateDesc = {}

--- ========================================================= ---
---  Setup Options
--- ========================================================= ---
Sphere_InnerGauge:RegisterSphereOption({
	name = L["Inner Gauge"],
	desc = L["Change the Inner Gauge options."],
	type = 'group',
	args = {
		event = {
			name = L["Event Type"],
			desc = L["Choose which Event Type the Gauge monitors and shows."],
			type = 'text',
			validate = Sphere_InnerGauge_eventTypeValidate,
			validateDesc = eventTypeValidateDesc,
			get = function()
				return self.db.char.gaugeType
			end,
			set = "SetGaugeType",
		},
		unit = {
			name = L["Event Unit"],
			desc = L["Choose which Unit the Event is monitoring and showing."],
			type = 'text',
			validate = eventUnitValidate,
			validateDesc = eventUnitValidateDesc,
			get = function()
				return self.db.char.gaugeUnit
			end,
			set = "SetGaugeUnit",
		},
		custom = {
			name = L["Custom Event"],
			desc = L["Create a Custom Event for the Gauge to monitor and show."],
			type = "text",
			get = function()
				return self.db.char.gaugeCustom
			end,
			set = "SetCustomGauge",
			usage = L["DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0"],
		},
		marks = {
			name = L["Marks Number"],
			desc = L["Choose how many Marks to show on the Gauge."],
			type = 'text',
			validate = gaugeMarksValidate,
			validateDesc = gaugeMarksValidateDesc,
			get = function()
				return self.db.char.gaugeMarks
			end,
			set = "SetGaugeMarks",
		},
		marksdarken = {
			name = L["Marks Darken"],
			desc = L["Choose whether or not to Darken the Marks."],
			type = 'toggle',
			get = function()
				return self.db.char.gaugeMarksDarken
			end,
			set = function() 
				self.db.char.gaugeMarksDarken = not(self.db.char.gaugeMarksDarken);
				self:SetGaugeMarksDarken();
			end,
		},
		animate = {
			name = L["Animate Gauge"],
			desc = L["Animate the Gauge (Glide)."],
			type = 'toggle',
			get = function()
				return self.db.char.animate
			end,
			set = function() 
				self.db.char.animate = not(self.db.char.animate);
			end,
		},
	}

})

--- ========================================================= ---
---  On Enable
--- ========================================================= ---
function Sphere_InnerGauge:OnEnable()
	self:ScheduleRepeatingEvent("UpdateInnerGauge", self.UpdateGauge, 0.05, self)
	SphereInnerFill:Show();
	SphereInnerHalf:SetID(0);
	SphereInnerHalf:Show();
	SphereInnerCover:Show();
	SphereInnerBorder:Show();
	SphereInnerMarks:Show();
end

--- ========================================================= ---
---  On Disable
--- ========================================================= ---
function Sphere_InnerGauge:OnDisable()
	self:UnregisterAllEvents();
	SphereInnerFill:Hide();
	SphereInnerHalf:Hide();
	SphereInnerCover:Hide();
	SphereInnerBorder:Hide();
	SphereInnerMarks:Hide();
end

--- ========================================================= ---
---  Add Event Type
--- ========================================================= ---
function Sphere_InnerGauge:AddEventType(func, key, value)
	func[key] = value
end

--- ========================================================= ---
---  Remove Event Type
--- ========================================================= ---
function Sphere_InnerGauge:RemoveEventType(func, key)
	func[key] = nil
end

--- ========================================================= ---
--- Create Sphere Frames
--- ========================================================= ---
function Sphere_InnerGauge:CreateFrames()
	Sphere:CreateFrame("Button", "SphereInnerFill", SphereMainFrame, 4, "Interface\\Addons\\Sphere\\art\\innerfill")
	Sphere:CreateFrame("Button", "SphereInnerCover", SphereMainFrame, 6, "Interface\\Addons\\Sphere\\art\\innercover")
	Sphere:CreateFrame("Button", "SphereInnerHalf", SphereMainFrame, 6, "Interface\\Addons\\Sphere\\art\\innerfill")
	Sphere:CreateFrame("Button", "SphereInnerBorder", SphereMainFrame, 8, "Interface\\Addons\\Sphere\\art\\innercircle")
	Sphere:CreateFrame("Button", "SphereInnerMarks", SphereMainFrame, 8, "Interface\\Addons\\Sphere\\art\\innermarks4")
	
	Sphere:RotateTexture(SphereInnerHalf:GetNormalTexture(), 180, 0.5, 0.5)
	SphereInnerHalf:Hide();
end

--- ========================================================= ---
---  Set Gauge Type
--- ========================================================= ---
function Sphere_InnerGauge:SetGaugeType(value)
	if value then
		self.db.char.gaugeType = value
		self.db.char.gaugeCustom = eventType[value]
	else
		value = self.db.char.gaugeType
	end
	
	if value == "health" or value == "power" or value == "threat" or value == "experience" or value == "reputation" or value == "custom" or value == "none" then
		 self.db.char.useCustom = false
	else
		self.db.char.useCustom = true
	end
end

--- ========================================================= ---
---  Set Gauge Unit
--- ========================================================= ---
function Sphere_InnerGauge:SetGaugeUnit(value)
	if value then
		self.db.char.gaugeUnit = value
	else
		value = self.db.char.gaugeUnit
	end
end

--- ========================================================= ---
---  Set Custom Gauge
--- ========================================================= ---
function Sphere_InnerGauge:SetCustomGauge(value)
	if value then
		if self.db.char.gaugeType == "custom" then
			self.db.char.gaugeCustom = value
		else
			self.db.char.gaugeCustom = eventType[self.db.char.gaugeType]
		end
	else
		value = self.db.char.gaugeCustom
	end
end

--- ========================================================= ---
--- Set Gauge Marks
--- ========================================================= ---

function Sphere_InnerGauge:SetGaugeMarks(value)
	if value then
		self.db.char.gaugeMarks = value
	else
		value = self.db.char.gaugeMarks
	end
	
	if value == "none" then value = 0 end
	if value == "four" then value = 4 end
	if value == "five" then value = 5 end
	if value == "ten" then value = 10 end
	if value == "twenty" then value = 20 end
	
	if value ~= 0 then
		SphereInnerMarks:SetNormalTexture("Interface\\Addons\\Sphere\\art\\innermarks" .. value);
		SphereInnerMarks:Show();
	else
		SphereInnerMarks:Hide();
	end

end

--- ========================================================= ---
--- Set Gauge Darken Marks
--- ========================================================= ---
function Sphere_InnerGauge:SetGaugeMarksDarken()
	value = self.db.char.gaugeMarksDarken

	if value then
		SphereInnerMarks:SetAlpha(1);
	else
		SphereInnerMarks:SetAlpha(0.5);
	end

end

--- ========================================================= ---
--- Update Gauge
--- ========================================================= ---
function Sphere_InnerGauge:UpdateGauge()
	local sPercent = 0
	local r, g, b, a
	
	if self.db.char.useCustom then
		sPercent, r, g, b, a  = Sphere[self.db.char.gaugeType](Sphere)
	else
		if self.db.char.gaugeType == "power" then
			local powerType = DogTag:Evaluate(self.db.char.gaugeUnit, "[TypePower]")
			if powerType == "" then powerType = "Mana" end
			r, g, b, a = unpack(Sphere.db.char.colors[string.lower(powerType)])
		else
			r, g, b, a = unpack(Sphere.db.char.colors[self.db.char.gaugeType])
		end
		
		if tonumber(DogTag:Evaluate(self.db.char.gaugeUnit, self.db.char.gaugeCustom)) then
			sPercent = tonumber(DogTag:Evaluate(self.db.char.gaugeUnit, self.db.char.gaugeCustom))
		else
			sPercent = 0
		end
		
		if self.db.char.animate then
			if (start == nil) then
				start = sPercent;
			end
			if (start > sPercent) then
				start = start - 1;
				if (start < sPercent) then
					start = sPercent;
				end
			else
				start = start + 1;
				if (start > sPercent) then
					start = sPercent;
				end
			end
			sPercent = start;
		end
	end

	if (sPercent > 100) then
		sPercent = 100;
	end

	if (sPercent <= 0) then
		SphereInnerHalf:Hide();
		SphereInnerHalf:SetID(0);
		SphereInnerCover:Show();
		SphereInnerFill:SetAlpha(0);
	end

	if (sPercent > 0) then
		if ((sPercent >= 50) and (SphereInnerHalf:GetID() ~= 1)) then
			SphereInnerHalf:Show();
			SphereInnerHalf:SetID(1);
			SphereInnerCover:Hide();
		elseif ((sPercent < 50) and (SphereInnerHalf:GetID() == 1)) then
			SphereInnerHalf:Hide();
			SphereInnerHalf:SetID(0);
			SphereInnerCover:Show();
		end
		SphereInnerFill:SetAlpha(1);
	else
		SphereInnerFill:SetAlpha(0);
	end
	
	SphereInnerFill:GetNormalTexture():SetVertexColor(r, g, b, a);
	SphereInnerHalf:GetNormalTexture():SetVertexColor(r, g, b, a);
	
	Sphere:RotateTexture(SphereInnerFill:GetNormalTexture(), math.floor(sPercent * 3.6), 0.5, 0.5)
end